package cn.iocoder.yudao.module.wealth.service.receiptInvoice;

import cn.iocoder.yudao.framework.common.exception.ErrorCode;
import cn.iocoder.yudao.framework.common.exception.ServiceException;
import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.framework.mybatis.core.service.AbstractService;
import cn.iocoder.yudao.framework.mybatis.core.util.MyBatisUtils;
import cn.iocoder.yudao.framework.mybatis.core.vo.PageVO;
import cn.iocoder.yudao.framework.security.core.LoginUser;
import cn.iocoder.yudao.framework.security.core.util.SecurityFrameworkUtils;
import cn.iocoder.yudao.module.ecw.api.currency.CurrencyApi;
import cn.iocoder.yudao.module.ecw.api.currency.dto.CurrencyRespDTO;
import cn.iocoder.yudao.module.ecw.api.currency.dto.ExchangeRateRespDTO;
import cn.iocoder.yudao.module.wealth.convert.receiptInvoice.ReceiptInvoiceConvert;
import cn.iocoder.yudao.module.wealth.dal.dataobject.receipt.ReceiptDO;
import cn.iocoder.yudao.module.wealth.dal.dataobject.receiptInvoice.ReceiptInvoiceDO;
import cn.iocoder.yudao.module.wealth.dal.dataobject.receiptInvoiceLog.ReceiptInvoiceLogDO;
import cn.iocoder.yudao.module.wealth.dal.mysql.receipt.ReceiptMapper;
import cn.iocoder.yudao.module.wealth.dal.mysql.receiptInvoice.ReceiptInvoiceMapper;
import cn.iocoder.yudao.module.wealth.dal.mysql.receiptInvoiceLog.ReceiptInvoiceLogMapper;
import cn.iocoder.yudao.module.wealth.enums.InvoicingLinkEnum;
import cn.iocoder.yudao.module.wealth.enums.InvoicingStatusEnum;
import cn.iocoder.yudao.module.wealth.service.receivable.ReceivableService;
import cn.iocoder.yudao.module.wealth.util.WealthGenCodeUtils;
import cn.iocoder.yudao.module.wealth.vo.receiptInvoice.*;
import cn.iocoder.yudao.module.wealth.vo.receivable.ReceivableBackVO;
import cn.iocoder.yudao.module.wealth.vo.receivable.WealthMoneyAmountVO;
import com.baomidou.mybatisplus.core.metadata.IPage;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;
import java.util.stream.Collectors;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.wealth.enums.ErrorCodeConstants.RECEIPT_NOT_EXISTS;
import static cn.iocoder.yudao.module.wealth.enums.ErrorCodeConstants.RECEIPT_VERIFICATION;
import static cn.iocoder.yudao.module.wealth.enums.InvoiceDataStatusEnum.REJECT;
import static cn.iocoder.yudao.module.wealth.enums.InvoiceDataStatusEnum.SUBMIT;
import static cn.iocoder.yudao.module.wealth.enums.InvoicingStatusEnum.CANCELLED;
import static cn.iocoder.yudao.module.wealth.enums.InvoicingStatusEnum.NO_RECORD;
import static cn.iocoder.yudao.module.wealth.enums.ReceiptStatusEnum.*;

/**
 * 收款单开票信息 Service 实现类
 *
 * @author 系统管理员
 */
@Service
@Validated
public class ReceiptInvoiceServiceImpl extends AbstractService<ReceiptInvoiceMapper, ReceiptInvoiceDO> implements ReceiptInvoiceService {

    @Resource
    private ReceiptInvoiceMapper receiptInvoiceMapper;
    @Resource
    private ReceiptInvoiceLogMapper receiptInvoiceLogMapper;
    @Resource
    private ReceiptMapper receiptMapper;
    @Resource
    private ReceivableService receivableService;
    @Resource
    private WealthGenCodeUtils wealthGenCodeUtils;
    @Resource
    private CurrencyApi currencyApi;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Long createReceiptInvoice(ReceiptInvoiceCreateReqVO createReqVO) {
        // 查询收款单
        ReceiptDO receiptDO = receiptMapper.selectById(createReqVO.getReceiptId());
        if (receiptDO == null) {
            throw exception(RECEIPT_NOT_EXISTS);
        }
        // 校验收款单状态
        checkReceiptForCreateInvoice(receiptDO);
        // 创建开票资料
        ReceiptInvoiceDO receiptInvoiceDO = ReceiptInvoiceConvert.INSTANCE.convert(createReqVO);
        receiptInvoiceDO.setInvoiceNo(wealthGenCodeUtils.generateInvoiceCode());
        receiptInvoiceDO.setReceiptId(receiptDO.getId());
        receiptInvoiceDO.setReceiptNo(receiptDO.getReceiptNo());
        receiptInvoiceDO.setOrderId(receiptDO.getOrderId());
        receiptInvoiceDO.setOrderNo(receiptDO.getOrderNo());
        LoginUser loginUser = SecurityFrameworkUtils.getLoginUser();
        receiptInvoiceDO.setApplyUserId(loginUser.getId());
        receiptInvoiceDO.setApplyUserName(loginUser.getUsername());
        receiptInvoiceMapper.insert(receiptInvoiceDO);
        // 更新收款单状态
        receiptDO.setInvoiceId(receiptInvoiceDO.getId());
        receiptDO.setInvoicingStatus(InvoicingStatusEnum.WAIT.getValue());
        receiptDO.setInvoiceDataStatus(SUBMIT.getValue());
        receiptMapper.updateById(receiptDO);
        // 记录日志 TODO 申请开票时没备注？
        recordLog(receiptInvoiceDO.getId(), InvoicingLinkEnum.APPLY_INVOICE, createReqVO.getRemark(), null);
        return receiptInvoiceDO.getId();
    }

    private void checkReceiptForCreateInvoice(ReceiptDO receiptDO) {
        //【申请开票】按钮：新增申请开票按钮，当收款单状态≠草稿且≠待审批且≠审批拒绝，开票状态=已取消或者无申请记录，按钮显示且可点击，跳转申请开票页面
        if (Objects.equals(receiptDO.getState(), DRAFT.getValue()) || Objects.equals(receiptDO.getState(), APPROVE_ING.getValue()) || Objects.equals(receiptDO.getState(), APPROVE_FAIL.getValue())) {
            throw exception(RECEIPT_VERIFICATION);
        }
        if (!Objects.equals(receiptDO.getInvoicingStatus(), CANCELLED.getValue()) && !Objects.equals(receiptDO.getInvoicingStatus(), NO_RECORD.getValue())) {
            throw exception(RECEIPT_VERIFICATION);
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void updateReceiptInvoiceData(ReceiptInvoiceUpdateDataReqVO updateReqVO) {
        // 校验存在
        this.validateReceiptInvoiceExists(updateReqVO.getId());
        // TODO 校验收款单开票状态为待开票
        // 更新开票资料
        ReceiptInvoiceDO updateObj = ReceiptInvoiceConvert.INSTANCE.convert(updateReqVO);
        receiptInvoiceMapper.updateById(updateObj);
        // 记录日志
        recordLog(updateReqVO.getId(), InvoicingLinkEnum.MODIFY_INVOICE_INFO, updateReqVO.getInvoiceRemark(), null);
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void updateReceiptInvoiceInfo(ReceiptInvoiceUpdateInfoReqVO updateReqVO) {
        this.validateReceiptInvoiceExists(updateReqVO.getId());
        ReceiptInvoiceDO updateObj = ReceiptInvoiceConvert.INSTANCE.convert(updateReqVO);
        receiptInvoiceMapper.updateById(updateObj);
        // 记录日志
        recordLog(updateReqVO.getId(), InvoicingLinkEnum.MODIFY_INVOICE_INFO, updateReqVO.getInvoiceInfoRemark(), null);
    }

    private void validateReceiptInvoiceExists(Long id) {
        if (receiptInvoiceMapper.selectById(id) == null) {
            throw exception(new ErrorCode(400, "收款单开票信息不存在，id=" + id));
        }
    }

    @Override
    public ReceiptInvoiceDetailVO getReceiptInvoiceDetail(Long id) {
        ReceiptInvoiceDetailVO receiptInvoiceDetail = receiptInvoiceMapper.getReceiptInvoiceDetail(id);
        if (receiptInvoiceDetail == null) {
            return null;
        }
        List<ReceivableBackVO> receivableList = receivableService.getReceivableListByReceiptId(receiptInvoiceDetail.getReceiptId());
        receiptInvoiceDetail.setReceivableList(receivableList);
        // 根据应收明细统计收款单信息
        statisticsReceiptInfo(receiptInvoiceDetail, receivableList);

        // 查询日志
        List<ReceiptInvoiceLogDO> invoiceLogDOS = receiptInvoiceLogMapper.selectList("invoice_id", id);
        receiptInvoiceDetail.setInvoiceLog(ReceiptInvoiceConvert.INSTANCE.logDOtoLogVOList(invoiceLogDOS));
        return receiptInvoiceDetail;
    }

    private void statisticsReceiptInfo(ReceiptInvoiceDetailVO receiptInvoiceDetail, List<ReceivableBackVO> receivableList) {
        // 应收明细按币种分组，统计每个币种的应收总金额和核销总金额
        Map<Long, List<ReceivableBackVO>> receiptAmountMap = receivableList.stream().collect(Collectors.groupingBy(ReceivableBackVO::getCurrencyId));
        List<WealthMoneyAmountVO> receiptAmountList = new ArrayList<>();
        Map<Integer, CurrencyRespDTO> currencyMap = currencyApi.getAllCurrency();
        for (Map.Entry<Long, List<ReceivableBackVO>> entry : receiptAmountMap.entrySet()) {
            Long currencyId = entry.getKey();
            List<ReceivableBackVO> currencyReceivableList = entry.getValue();
            WealthMoneyAmountVO wealthMoneyAmountVO = new WealthMoneyAmountVO();
            wealthMoneyAmountVO.setCurrencyId(currencyId.intValue());
            wealthMoneyAmountVO.setCurrencySymbol(currencyMap.get(currencyId.intValue()).getFuhao());
            wealthMoneyAmountVO.setCurrencyName(currencyMap.get(currencyId.intValue()).getTitleZh());
            wealthMoneyAmountVO.setTotalAmount(currencyReceivableList.stream().map(ReceivableBackVO::getTotalAmount).reduce(BigDecimal.ZERO, BigDecimal::add));
            wealthMoneyAmountVO.setNotWriteOffAmount(currencyReceivableList.stream().map(ReceivableBackVO::getNotWriteOffAmount).reduce(BigDecimal.ZERO, BigDecimal::add));
            wealthMoneyAmountVO.setWriteOffAmount(currencyReceivableList.stream().map(ReceivableBackVO::getWriteOffAmount).reduce(BigDecimal.ZERO, BigDecimal::add));
            wealthMoneyAmountVO.setReceivableTotalAmount(currencyReceivableList.stream().map(vo -> vo.getTaxAmount().subtract(vo.getDiscountTotal())).reduce(BigDecimal.ZERO, BigDecimal::add));
            receiptAmountList.add(wealthMoneyAmountVO);
        }
        receiptInvoiceDetail.setReceiptAmount(receiptAmountList);
        receiptInvoiceDetail.setReceivableTotalAmountUsd(BigDecimal.ZERO);
        receiptInvoiceDetail.setWriteOffAmountUsd(BigDecimal.ZERO);
        for (WealthMoneyAmountVO wealthMoneyAmountVO : receiptAmountList) {
            ExchangeRateRespDTO rate = currencyApi.getCurrencyRateByCode(wealthMoneyAmountVO.getCurrencySymbol(), "$");
            receiptInvoiceDetail.setReceivableTotalAmountUsd(receiptInvoiceDetail.getReceivableTotalAmountUsd().add(wealthMoneyAmountVO.getReceivableTotalAmount().multiply(rate.getCurrencyRate())));
            receiptInvoiceDetail.setWriteOffAmountUsd(receiptInvoiceDetail.getWriteOffAmountUsd().add(wealthMoneyAmountVO.getWriteOffAmount().multiply(rate.getCurrencyRate())));
        }
        receiptInvoiceDetail.setWriteOffProportion(receiptInvoiceDetail.getWriteOffAmountUsd().divide(receiptInvoiceDetail.getReceivableTotalAmountUsd(), 4, RoundingMode.HALF_UP).toString());
        receiptInvoiceDetail.setCurrencySymbol(currencyMap.get(receiptInvoiceDetail.getCurrencyId().intValue()).getFuhao());
    }

    @Override
    public List<ReceiptInvoiceDO> getReceiptInvoiceList(Collection<Long> ids) {
        return receiptInvoiceMapper.selectBatchIds(ids);
    }

    @Override
    public PageResult<ReceiptInvoicePageBackVO> getReceiptInvoicePage(ReceiptInvoiceQueryVO query, PageVO page) {
        IPage<ReceiptInvoicePageBackVO> mpPage = MyBatisUtils.buildPage(page);
        receiptInvoiceMapper.getReceiptInvoicePage(mpPage, query);
        return PageResult.of(mpPage);
    }

    @Override
    public List<ReceiptInvoiceDO> getReceiptInvoiceList(ReceiptInvoiceQueryVO query) {
        return receiptInvoiceMapper.selectList(query);
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Boolean rejectInvoiceInfo(ReceiptInvoiceOperateReqVO reqVO) {
        ReceiptInvoiceDO receiptInvoiceDO = receiptInvoiceMapper.selectById(reqVO.getId());
        if (receiptInvoiceDO == null) {
            throw new ServiceException(new ErrorCode(400, "发票不存在"));
        }
        ReceiptDO receiptDO = new ReceiptDO();
        receiptDO.setId(receiptInvoiceDO.getReceiptId());
        receiptDO.setInvoiceDataStatus(REJECT.getValue());
        receiptMapper.updateById(receiptDO);
        recordLog(receiptInvoiceDO.getReceiptId(), InvoicingLinkEnum.REJECT_INVOICE_DATA, reqVO.getRemark(), null);
        // TODO 通知抄送人
        return true;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Boolean cancelInvoice(ReceiptInvoiceOperateReqVO reqVO) {
        ReceiptInvoiceDO receiptInvoiceDO = receiptInvoiceMapper.selectById(reqVO.getId());
        if (receiptInvoiceDO == null) {
            throw new ServiceException(new ErrorCode(400, "发票不存在"));
        }
        ReceiptDO receiptDO = new ReceiptDO();
        receiptDO.setId(receiptInvoiceDO.getReceiptId());
        receiptDO.setInvoicingStatus(CANCELLED.getValue());
        receiptMapper.updateById(receiptDO);
        recordLog(receiptInvoiceDO.getReceiptId(), InvoicingLinkEnum.CANCEL_INVOICE, reqVO.getRemark(), null);
        // TODO 通知抄送人
        return true;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Boolean deleteInvoiceInfo(ReceiptInvoiceOperateReqVO reqVO) {
        ReceiptInvoiceDO receiptInvoiceDO = receiptInvoiceMapper.selectById(reqVO.getId());
        if (receiptInvoiceDO == null) {
            throw new ServiceException(new ErrorCode(400, "发票不存在"));
        }
        // 清空发票号码，开票类型，开票金额，开票人，开票时间
        receiptInvoiceDO.setInvoiceNumber(null);
        receiptInvoiceDO.setInvoiceTypeId(null);
        receiptInvoiceDO.setInvoiceMoney(null);
        receiptInvoiceDO.setCurrencyId(null);
        receiptInvoiceDO.setInvoicerId(null);
        receiptInvoiceDO.setInvoiceTime(null);
        receiptInvoiceDO.setInvoiceInfoSaveTime(new Date());
        receiptInvoiceMapper.updateById(receiptInvoiceDO);

        ReceiptDO receiptDO = new ReceiptDO();
        receiptDO.setId(receiptInvoiceDO.getReceiptId());
        receiptDO.setInvoicingStatus(InvoicingStatusEnum.WAIT.getValue());
        receiptMapper.updateById(receiptDO);
        recordLog(receiptInvoiceDO.getReceiptId(), InvoicingLinkEnum.DELETE_INVOICE_INFO, reqVO.getRemark(), null);
        // TODO 通知抄送人
        return null;
    }

    /**
     * 记录日志
     * @param invoiceId 发票id
     * @param linkEnum 环节
     * @param remark 备注
     * @param bpmStatus 审批流程
     */
    private void recordLog(Long invoiceId, InvoicingLinkEnum linkEnum, String remark, Integer bpmStatus) {
        LoginUser user = SecurityFrameworkUtils.getLoginUser();
        String username = "";
        if (user != null && user.getUsername() != null) {
            username = user.getUsername();
        }
        ReceiptInvoiceLogDO receiptInvoiceLogDO = ReceiptInvoiceLogDO.builder()
                .invoiceId(invoiceId)
                .operator(username)
                .remark(remark)
                .invoicingLink(linkEnum.getValue())
                .bpmStatus(bpmStatus)
                .createTime(new Date())
                .build();
        receiptInvoiceLogMapper.insert(receiptInvoiceLogDO);
    }
}
